;
;*******************************************************************
; 	BMON1048.H                    Date 22.12.2010
;	Battery Monitor with PIC18F2520
;       by Anja
;       An_ja  @  http://www.mikrocontroller.net
;
;       V1: 02.03.2013
;       - comments added (Hex neutral changes only)
;*******************************************************************
;
; clock frequency around 1.0 MHz with internal R/C oscillator @ 5V
;
;*******************************************************************
;
Clock	equ	1000000       ; 1 MHz R/C clock
Clock4  equ	Clock / 4     ; command cycle = clock/4
;
;*******************************************************************************
;
;	PINS
;
;*******************************************************************************
;
#define LoadON  LATB,6        ; output to load
#define ExtRef  LATA,1        ; 
#define DP1     LATB,5        ; left decimal point on LCD
#define DP2     LATB,1        ; mid  decimal point on LCD
#define Key     PORTA,4       ; on/off key
;
;*******************************************************************************
;
;	MACROS
;
;*******************************************************************************
;
#define Cells   12              ; # number of cells
#define mV      * 256 * 51/1051/1235 ; 25,6V * 5K1/(100K+5K1)/VRef
#define VOn     Cells * 1250 mV ; high hysteresis
#define VOff    Cells * 1150 mV ; low  hysteresis
#define VOLT_SEC 10             ; check every 10 seconds
#define VDIFF   600 mV          ; voltage drop / minute
#define mA      * 1             ; 1 mA / count
#define CURR_MIN 10 mA          ; min current to display 
;
; clock for delay macros
;
#define ms	* (Clock4 / 1000)
#define	us	* Clock4 / 1000000
NumLoop equ	4		; 4 clocks/loop
;
Delay	macro	Cycles		;	
CYC	set	Cycles/NumLoop	;
    if	CYC <= 255		;
	movlw	CYC		;
	call	Delay8		;
    else
CYC	set	(CYC+255)/256	;
	movlw	CYC		;
	call	Delay16		;
    endif
	endm
;
;*******************************************************************
; Variablen
;*******************************************************************
;
;
    CBLOCK 0x00         ; 
;
	HZ60		; Counter 60Hz/second
	TEMP		; Temp
        OldState        ; debouncing of voltage
        Volt            ; VBat
	Flags		; common state flags (usage see at end of variables)
;
; math registers see MCHIP ANs
;
        AARGB0          ; MSB
        AARGB1          ;
        AARGB2          ;
        AARGB3          ; LSB
;
        BARGB0          ; MSB
        BARGB1          ; LSB
;
        TEMPB0          ; MSB
        TEMPB1          ; LSB
;
; BCD-Digits of LCD
;
        LCD0            ;
        LCD1            ;
        LCD2            ;
;
; storage for segments on port A-C
;
        SEGA            ;
        SEGB            ;
        SEGC            ;
;
; shift register for key debouncing
;
        KeyDeb          ;
;
	VoltCou		; counter 10 seconds to check dU/dt
	VoltIdx		; index for ring buffer
;
	VoltOld0	; ring buffer for dU/dt check
	VoltOld1	;
	VoltOld2	;
	VoltOld3	;
	VoltOld4	;
	VoltOld5	;
;
	Volt2		; old voltage for debouncing
;
	VoltH		; Voltage raw value H
        VoltL		;
;
	CurrH		; Current raw value H
	CurrL		;
; 
	Curr		; corrected current value
        CurrCou		;
;
	VoltCorrH	; voltage correction values from EEPROM
	VoltCorrL	;
	CurrCorrH	; current correction values from EEPROM
	CurrCorrL	;
;
    ENDC
;
;*******************************************************************
; Flag-Bits
;*******************************************************************
;
#define VoltOK	Flags,0 ; 1 = voltage ok to switch on
#define KeyOn   Flags,1 ; 1 = key state to switch on
#define CurrOK  Flags,2
;
;
; Register-Bits
;
#define _c	STATUS,C ; Carry des Statusreg
;
;*******************************************************************
; 
